# aes_gcm.py
import base64
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class AES256GCM:
    def __init__(self, key: bytes):
        """
        Initialize AES-256-GCM cipher with a key.

        Args:
            key: A 32-byte key required for AES-256 encryption/decryption

        Raises:
            ValueError: If key is not exactly 32 bytes
            AssertionError: If key is None
        """
        if key is not None and len(key) != 32:
            raise ValueError("Key must be 32 bytes")
        assert key is not None, "Key must be provided"
        self.key = key

    def encrypt(self, plaintext: str) -> str:
        """
        Encrypt a string using AES-256-GCM
        Args:
            plaintext: String to encrypt
        Returns:
            base64 strings of (ciphertext)_(iv)_(tag) joined by underscores
        """
        iv = os.urandom(12)
        cipher = Cipher(
            algorithms.AES(self.key), modes.GCM(iv), backend=default_backend()
        )
        encryptor = cipher.encryptor()
        ciphertext = encryptor.update(plaintext.encode()) + encryptor.finalize()

        return (
            f"{base64.b64encode(ciphertext).decode('utf-8')}_"
            f"{base64.b64encode(iv).decode('utf-8')}_"
            f"{base64.b64encode(encryptor.tag).decode('utf-8')}"
        )

    def decrypt(self, encrypted: str) -> str:
        """
        Decrypt a string using AES-256-GCM
        Args:
            encrypted: base64 strings of (ciphertext)_(iv)_(tag) joined by underscores
        Returns:
            Decrypted string
        """
        try:
            ciphertext_b64, iv_b64, tag_b64 = encrypted.split("_")
            ciphertext = base64.b64decode(ciphertext_b64)
            iv = base64.b64decode(iv_b64)
            tag = base64.b64decode(tag_b64)
        except (ValueError, base64.binascii.Error) as e:
            raise ValueError("Invalid encrypted format") from e

        cipher = Cipher(
            algorithms.AES(self.key), modes.GCM(iv, tag), backend=default_backend()
        )
        decryptor = cipher.decryptor()

        try:
            plaintext = decryptor.update(ciphertext) + decryptor.finalize()
            return plaintext.decode("utf-8")
        except Exception as e:
            raise ValueError("Decryption failed") from e



from io import BytesIO, IOBase

from smart_open import open

from ...s3.client import S3Client
from ..shared import OpenDALClient


class OpenDALS3Client(OpenDALClient):
    def __init__(
        self,
        bucket,
        region=None,
        endpoint=None,
        access_key_id=None,
        secret_access_key=None,
    ):
        """
        Initialize the OpenDALS3Client with AWS credentials and configuration.

        :param bucket: Name of the S3 bucket.
        :param region: AWS region name.
        :param endpoint: Custom S3 endpoint URL.
        :param access_key_id: AWS access key ID.
        :param secret_access_key: AWS secret access key.
        """
        self._bucket = bucket
        self._region = region
        self._endpoint = endpoint
        self._access_key_id = access_key_id
        self._secret_access_key = secret_access_key

        # Initialize S3 client with credentials
        credentials = None
        if access_key_id and secret_access_key:
            credentials = {
                "aws_access_key_id": access_key_id,
                "aws_secret_access_key": secret_access_key,
            }

        self.operator = S3Client(
            credentials=credentials,
            endpoint_url=endpoint,
            region_name=region,
        )

    def read(self, path: str) -> bytes:
        with self.operator.download_to_temporary_file(self._bucket, path) as temp_file:
            with open(temp_file, "rb") as f:
                return f.read()

    def presign_read(self, path: str, expiration: int) -> str:
        class PresignedUrl:
            def __init__(self, url):
                self.url = url

        return PresignedUrl(
            self.operator.generate_presigned_get_url(
                self._bucket, path, expiration=expiration
            )
        )

    def stat(self, path: str) -> dict:
        class Stat:
            def __init__(self, stat):
                self.stat = stat

            @property
            def content_length(self):
                return self.stat["ContentLength"]

        return Stat(self.operator.head_object(self._bucket, path))

    def open(self, path: str, mode: str = "rb") -> IOBase:
        return open(
            f"s3://{self._bucket}/{path}",  # noqa
            mode,
            transport_params={"client": self.operator.client},
        )

    def write(self, path: str, data: bytes):
        fileobj = BytesIO(data)
        self.operator.upload_fileobj(fileobj, self._bucket, path)

    def delete(self, path: str):
        self.operator.delete_object(self._bucket, path)

    @property
    def bucket(self):
        return self._bucket

    @property
    def region(self):
        return self._region

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def access_key_id(self):
        return self._access_key_id

    @property
    def secret_access_key(self):
        return self._secret_access_key
